/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                           BIAYSTOK 2018                               **
**                                                                        **
**                         Home Control System                            **
**                                                                        **
**                             HC_PROCO2                                  **
**                      Wersja z dnia : 2018.09.01                        **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/


#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>


#define false                          0
#define true                           1

#define LEDPort                        PORTB
#define LEDPortDirection               DDRB

#define StatusLed                      4
#define RS232RxLed                     1
#define RS232TxLed                     3
#define RS485RxLed                     0
#define RS485TxLed                     2

#define TimerCounterDelayLimit         2048
#define StandardLEDPulseDelay          832

#define RxTxPort                       PORTD
#define RxTxDirectionPort              DDRD
#define RxTxDirPin                     4              // PORTD . 4

#define TL16C450ResetPort              PORTD
#define TL16C450ResetDirectionPort     DDRD
#define TL16C450ResetPin               5              // PORTD . 5

              /* adresy poszczegolnych rejestrow UART 16450 (16550) */
#define DataRegister                   0x8000
#define InterruptEnRegister            0x8100
#define InterruptIdRegister            0x8200
#define LineControlRegister            0x8300
#define ModemControlRegister           0x8400
#define LineStatusRegister             0x8500
#define ModemStatusRegister            0x8600
#define ScratchRegister                0x8700

#define nop() __asm__ __volatile__ ("nop")

// USART BAUD Rate Generator fosc=7.3728 MHz
#define RS485Speed9600bps              47
#define RS485Speed19200bps             23
#define RS485Speed38400bps             11
#define RS485Speed115200bps            3

#define RS232Serial38400bps            3
#define RS232Serial19200bps            6
#define RS232Serial9600bps             12

#define IdleState                      0
#define BusyState                      1

#define TXDBuffSize                    128
#define RXDBuffSize                    16

typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t State ;
                 uint8_t Buffer [ TXDBuffSize ] ;
               } TXDRecT ;
typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t Buffer [ RXDBuffSize ] ;
               } RXDRecT ;


#define CommandBufferSize              64
typedef uint8_t CommandBufferType [ CommandBufferSize + 1 ] ;


volatile static uint8_t TimeEvent ;
static volatile TXDRecT RS232TXDRec ;
static volatile RXDRecT RS232RXDRec ;
static volatile TXDRecT RS485TXDRec ;
static volatile RXDRecT RS485RXDRec ;
static uint16_t RS232RxSerialLEDDelay ;
static uint16_t RS232TxSerialLEDDelay ;
static uint16_t RS485RxSerialLEDDelay ;
static uint16_t RS485TxSerialLEDDelay ;
static CommandBufferType RS232Command ;
static uint16_t TimerCounter ;
static uint16_t RS232CommandIndex ;
static uint16_t RS485CommandIndex ;


static uint8_t DeviceName [ ] PROGMEM               = "\r\nHC-PROCO2 - transparentny procesor komunikacyjny\r\n" ;



typedef enum { LifeStatusLamp ,
               RS232RxLamp ,
               RS485RxLamp ,
               RS232TxLamp ,
               RS485TxLamp ,
             } LampSinalT ;


static void LEDOn ( LampSinalT LEDSignal )
{
  /*----------------------------------------------------------------*/
  switch ( LEDSignal )
  {
    case LifeStatusLamp :
      LEDPort &= ~ ( 1 << StatusLed ) ;
      break ;
    case RS485RxLamp    :
      LEDPort &= ~ ( 1 << RS485RxLed ) ;
      break ;
    case RS485TxLamp    :
      LEDPort &= ~ ( 1 << RS485TxLed ) ;
      break ;
    case RS232RxLamp    :
      LEDPort &= ~ ( 1 << RS232RxLed ) ;
      break ;
    case RS232TxLamp    :
      LEDPort &= ~ ( 1 << RS232TxLed ) ;
      break ;
  } /* switch */ ;
} /* LEDOn */



static void LEDOff ( LampSinalT LEDSignal )
{
  /*----------------------------------------------------------------*/
  switch ( LEDSignal )
  {
    case LifeStatusLamp :
      LEDPort |= ( 1 << StatusLed ) ;
      break ;
    case RS485RxLamp    :
      LEDPort |= ( 1 << RS485RxLed ) ;
      break ;
    case RS485TxLamp    :
      LEDPort |= ( 1 << RS485TxLed ) ;
      break ;
    case RS232RxLamp    :
      LEDPort |= ( 1 << RS232RxLed ) ;
      break ;
    case RS232TxLamp    :
      LEDPort |= ( 1 << RS232TxLed ) ;
      break ;
  } /* switch */ ;
} /* LEDOff */


ISR ( TIMER0_OVF_vect )
{
  /*----------------------------------------------------------------*/
  TimerCounter ++ ;
  if ( TimerCounter > TimerCounterDelayLimit )
  {
    TimerCounter = 0 ;
    TimeEvent = true ;
  } /* if */ ;
  if ( RS232RxSerialLEDDelay )
  {
    RS232RxSerialLEDDelay -- ;
    if ( ! RS232RxSerialLEDDelay )
    {
      LEDOff ( RS232RxLamp ) ;
    } /* if */ ;
  } /* if */ ;
  if ( RS232TxSerialLEDDelay )
  {
    RS232TxSerialLEDDelay -- ;
    if ( ! RS232TxSerialLEDDelay )
    {
      LEDOff ( RS232TxLamp ) ;
    } /* if */ ;
  } /* if */ ;
  if ( RS485RxSerialLEDDelay )
  {
    RS485RxSerialLEDDelay -- ;
    if ( ! RS485RxSerialLEDDelay )
    {
      LEDOff ( RS485RxLamp ) ;
    } /* if */ ;
  } /* if */ ;
  if ( RS485TxSerialLEDDelay )
  {
    RS485TxSerialLEDDelay -- ;
    if ( ! RS485TxSerialLEDDelay )
    {
      LEDOff ( RS485TxLamp ) ;
    } /* if */ ;
  } /* if */ ;
} /* TIMER0_OVF_vect */


static void StartRS232RxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOn ( RS232RxLamp ) ;
  RS232RxSerialLEDDelay = StandardLEDPulseDelay ;
} /* StartRS232RxLED */


static void StartRS232TxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOn ( RS232TxLamp ) ;
  RS232TxSerialLEDDelay = StandardLEDPulseDelay ;
} /* StartRS232TxLED */


static void StartRS485RxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOn ( RS485RxLamp ) ;
  RS485RxSerialLEDDelay = StandardLEDPulseDelay ;
} /* StartRS485RxLED */


static void StartRS485TxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOn ( RS485TxLamp ) ;
  RS485TxSerialLEDDelay = StandardLEDPulseDelay ;
} /* StartRS485TxLED */


static void RS485ReceiveMode ( void )
{
  /*----------------------------------------------------------------*/
  RxTxPort &= ~ ( 1 << RxTxDirPin ) ;
} /* RS485ReceiveMode */


static void RS485TransmitMode ( void )
{
  /*----------------------------------------------------------------*/
  RxTxPort |= ( 1 << RxTxDirPin ) ;
} /* RS485TransmitMode */


ISR ( INT0_vect )
{
  volatile uint8_t * Port ;
  uint8_t IIR ;
  uint8_t UARTStatus ;
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  Port = ( uint8_t * ) InterruptIdRegister ;
  IIR = ( * Port ) & 0x07 ;
  if ( ( IIR & 0x01 ) == 0 )
  {
    IIR = IIR >> 1 ;
    switch ( IIR )
    {
      case 0             :   /* 0 0 - Modem status */
        Port = ( uint8_t * ) ModemStatusRegister ;
        UARTStatus = * Port ;
        break ;
      case 1             :   /* 0 1 - transmitter */
        if ( RS232TXDRec . CycFr == RS232TXDRec . CycTo )
        {
          RS232TXDRec . State = IdleState ;
        } /* if ... */
        else
        {
          Port = ( uint8_t * ) DataRegister ;
          * Port = RS232TXDRec . Buffer [ RS232TXDRec . CycFr ] ;
          RS232TXDRec . CycFr ++ ;
          if ( RS232TXDRec . CycFr >= TXDBuffSize )
            RS232TXDRec . CycFr = 0 ;
          StartRS232TxLED ( ) ;
        } /* if ... else */ ;
        break ;
      case 2             :   /* 1 0 - reciever */
        Port = ( uint8_t * ) DataRegister ;
        Data = * Port ;
        RS232RXDRec . Buffer [ RS232RXDRec . CycTo ] = Data ;
        RS232RXDRec . CycTo ++ ;
        if ( RS232RXDRec . CycTo >= RXDBuffSize )
          RS232RXDRec . CycTo = 0 ;
        StartRS232RxLED ( ) ;
        break ;
      case 3             : /* error */
        Port = ( uint8_t * ) LineStatusRegister ;
        UARTStatus = * Port ;
        break ;
    } /* switch */ ;
  } /* if */ ;
} /* INT0_vect */


ISR ( UART_RX_vect )
{
  uint8_t RecCh ;
  /*----------------------------------------------------------------*/
  RecCh = UDR ;
  RS485RXDRec . Buffer [ RS485RXDRec . CycTo ] = RecCh ;
  RS485RXDRec . CycTo ++ ;
  if ( RS485RXDRec . CycTo >= RXDBuffSize )
    RS485RXDRec . CycTo = 0 ;
  StartRS485RxLED ( ) ;
} /* UART_RX_vect */


ISR ( UART_TX_vect )
{
  /*----------------------------------------------------------------*/
  if ( RS485TXDRec . CycFr == RS485TXDRec . CycTo )
  {
    RS485TXDRec . State = IdleState ;
    RS485ReceiveMode ( ) ;
  } /* if ... */
  else
  {
    UDR = RS485TXDRec . Buffer [ RS485TXDRec . CycFr ] ;
    RS485TXDRec . CycFr ++ ;
    if ( RS485TXDRec . CycFr >= TXDBuffSize )
      RS485TXDRec . CycFr = 0 ;
    StartRS485TxLED ( ) ;
  } /* if ... else */ ;
} /* UART_TX_vect */


static void RS232SendSerial ( uint8_t SerialCh )
{
  volatile uint8_t * Port ;
  uint8_t NextInx ;
  /*----------------------------------------------------------------*/
  cli ( ) ;
  StartRS232TxLED ( ) ;
  if ( RS232TXDRec . State == IdleState )
  {
    RS232TXDRec . State = BusyState ;
    Port = ( uint8_t * ) DataRegister ;
    * Port = SerialCh ;
    sei ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      NextInx = RS232TXDRec . CycTo + 1 ;
      if ( NextInx >= TXDBuffSize )
        NextInx = 0 ;
      if ( NextInx != RS232TXDRec . CycFr )
        break ;
      sei ( ) ;
      nop ( ) ;
      nop ( ) ;
      nop ( ) ;
      wdt_reset ( ) ;
      cli ( ) ;
    } /* for */ ;
    RS232TXDRec . Buffer [ RS232TXDRec . CycTo ] = SerialCh ;
    RS232TXDRec . CycTo ++ ;
    if ( RS232TXDRec . CycTo >= TXDBuffSize )
      RS232TXDRec . CycTo = 0 ;
    sei ( ) ;
  } /* if ... else */ ;
} /* RS232SendSerial */



static void RS232SendSerialFlashTxt ( uint16_t String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = pgm_read_byte_near ( String ) ;
    if ( Data )
      RS232SendSerial ( Data ) ;
    else
      break ;
    String ++ ;
  } /* for */ ;
} /* RS232SendSerialFlashTxt */


static uint8_t RS232SerialDataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( RS232RXDRec . CycFr == RS232RXDRec . CycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* RS232SerialDataPresent */


static uint8_t RS232GetSerial ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = RS232RXDRec . Buffer [ RS232RXDRec . CycFr ] ;
  RS232RXDRec . CycFr ++ ;
  if ( RS232RXDRec . CycFr >= RXDBuffSize )
    RS232RXDRec . CycFr = 0 ;
  return ( Ch ) ;
} /* RS232GetSerial */


static void RS485SendSerial ( uint8_t SerialCh )
{
  uint8_t NextInx ;
  /*----------------------------------------------------------------*/
  cli ( ) ;
  StartRS485TxLED ( ) ;
  if ( RS485TXDRec . State == IdleState )
  {
    RS485TransmitMode ( ) ;
    RS485TXDRec . State = BusyState ;
    UDR = SerialCh ;
    sei ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      NextInx = RS485TXDRec . CycTo + 1 ;
      if ( NextInx >= TXDBuffSize )
        NextInx = 0 ;
      if ( NextInx != RS485TXDRec . CycFr )
        break ;
      sei ( ) ;
      nop ( ) ;
      nop ( ) ;
      nop ( ) ;
      wdt_reset ( ) ;
      cli ( ) ;
    } /* for */ ;
    RS485TXDRec . Buffer [ RS485TXDRec . CycTo ] = SerialCh ;
    RS485TXDRec . CycTo ++ ;
    if ( RS485TXDRec . CycTo >= TXDBuffSize )
      RS485TXDRec . CycTo = 0 ;
    sei ( ) ;
  } /* if ... else */ ;
} /* RS485SendSerial */


static uint8_t RS485SerialDataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( RS485RXDRec . CycFr == RS485RXDRec . CycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* RS485SerialDataPresent */


static uint8_t RS485GetSerial ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = RS485RXDRec . Buffer [ RS485RXDRec . CycFr ] ;
  RS485RXDRec . CycFr ++ ;
  if ( RS485RXDRec . CycFr >= RXDBuffSize )
    RS485RXDRec . CycFr = 0 ;
  return ( Ch ) ;
} /* RS485GetSerial */


static void RS485UARTHardwareInit ( uint8_t RS485SerialSpeed )
{
  /*----------------------------------------------------------------*/
  UBRRH = 0 ;
  UBRRL = RS485SerialSpeed ;
  UCSRB = ( 1 << RXCIE ) | ( 1 << TXCIE ) | ( 1 << RXEN ) | ( 1 << TXEN ) ;
  UCSRC = ( 1 << URSEL ) | ( 1 << UCSZ1 ) | ( 1 << UCSZ0 ) ;
} /* RS485UARTHardwareInit */


static void RS232UARTHardwareInit ( uint8_t RS232SerialSpeed )
{
  volatile uint8_t * Port ;
  uint8_t Data ;
  uint8_t UARTMode ;
  /*----------------------------------------------------------------*/
  UARTMode = 0x03 ;             /* Tryb: N81 */
  Port = ( uint8_t * ) LineControlRegister ;
  * Port = 0x080 ;              /* ustawienie w UART rozszerzenia adresowego do zaprogramowania predkosci */
  Port = ( uint8_t * ) DataRegister ;
  * Port = RS232SerialSpeed ;   /* mlodszy bajt podzielnika predkosci */
  Port = ( uint8_t * ) InterruptEnRegister ;
  * Port = 0 ;                  /* starszy bajt podzielnika predkosci */
  Port = ( uint8_t * ) LineControlRegister ;
  * Port = UARTMode ;           /* zaprogramowania trybu transmisji (N81) */
  Port = ( uint8_t * ) ModemControlRegister ;
  * Port = 0x00B ;              /* zaprogramowania stanu linii modemowych-> wszystkie do stanu aktywnego */
  Port = ( uint8_t * ) LineStatusRegister ;
  Data = * Port ;               /* wygarniecie smieci z roznych rejestrow */
  Port = ( uint8_t * ) ModemControlRegister ;
  Data = * Port ;               /* wygarniecie smieci z roznych rejestrow */
  Port = ( uint8_t * ) DataRegister ;
  Data = * Port ;               /* wygarniecie smieci z roznych rejestrow */
  Data = * Port ;               /* wygarniecie smieci z roznych rejestrow */
  Port = ( uint8_t * ) InterruptIdRegister ;
  Data = * Port ;               /* wygarniecie smieci z roznych rejestrow */
  Port = ( uint8_t * ) InterruptEnRegister ;
  * Port = ( uint8_t ) 0x03 ;   /* zezwolenie dla UART na przerwania od transmisji: nadajnik + odbiornik */
} /* RS232UARTHardwareInit */


static void HardwareInit ( void )
{
  /*----------------------------------------------------------------*/
  TIMSK = ( 1 << TOIE0 ) ;
  TCCR0 = ( 1 << CS01 ) ; // CS02=0 CS01=1 CS00=0 : F clk div 8
  TCNT0 = 0 ;
  MCUCR = ( 1 << SRE ) | ( 1 << SRW10 ) | ( 1 << ISC01 ) ; // ISC01=1 ISC00=0 -> falling edge
  GICR = ( 1 << INT0 ) ;
} /* HardwareInit */


static void SoftwareInit ( void )
{
  /*----------------------------------------------------------------*/
  RS485RxSerialLEDDelay = 0 ;
  RS485TxSerialLEDDelay = 0 ;
  RS232RxSerialLEDDelay = 0 ;
  RS232TxSerialLEDDelay = 0 ;
  RS232RXDRec . CycFr = 0 ;
  RS232RXDRec . CycTo = 0 ;
  RS232TXDRec . CycFr = 0 ;
  RS232TXDRec . CycTo = 0 ;
  RS232TXDRec . State = IdleState ;
  RS485RXDRec . CycFr = 0 ;
  RS485RXDRec . CycTo = 0 ;
  RS485TXDRec . CycFr = 0 ;
  RS485TXDRec . CycTo = 0 ;
  RS485TXDRec . State = IdleState ;
  TimerCounter = 0 ;
  TimeEvent = false ;
  RS232CommandIndex = 0 ;
  RS485CommandIndex = 0 ;
} /* SoftwareInit */


static void RetransmitToRS485 ( void )
{
  uint8_t Ch ;
  uint16_t Loop ;
  /*----------------------------------------------------------------*/
  if ( RS232CommandIndex )
  {
    for ( Loop = 0 ; Loop < RS232CommandIndex ; Loop ++ )
    {
      Ch = RS232Command [ Loop ] ;
      RS485SendSerial ( Ch ) ;
    } /* for */ ;
  } /* if */ ;
} /* RetransmitToRS485 */


static void RS232ProcessData ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( RS232CommandIndex < CommandBufferSize )
  {
    RS232Command [ RS232CommandIndex ] = Ch ;
    RS232CommandIndex ++ ;
  } /* if */ ;
  if ( Ch == 0x0D )
  {
    RetransmitToRS485 ( ) ;
    RS232CommandIndex = 0 ;
    RS232Command [ 0 ] = 0 ;
  } /* if */ ;
} /* RS232ProcessData */


static void OpenDelay ( void )
{
  uint16_t Loop ;
  /*-------------------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < 20000 ; Loop ++ )
  {
    if ( TimeEvent )
    {
      break ;
    } /* if */ ;
    wdt_reset ( ) ;
  } /* for */ ;
} /* OpenDelay */


static void EnvirInit ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = 0xFF ;
  LEDPortDirection |= Bits ;
  LEDPort |= Bits ;
  Bits = ( 1 << RxTxDirPin ) ;
  RxTxDirectionPort |= Bits ;
  RS485ReceiveMode ( ) ;
  TL16C450ResetDirectionPort |= ( 1 << TL16C450ResetPin ) ;
  TL16C450ResetPort &= ~ ( 1 << TL16C450ResetPin ) ;
} /* EnvirInit */


static void LEDTestStart ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = ( 1 << StatusLed ) | ( 1 << RS232RxLed ) | ( 1 << RS232TxLed ) | ( 1 << RS485RxLed ) | ( 1 << RS485TxLed ) ;
  LEDPort &= ~ Bits ;
} /* LEDTestStart */


static void LEDTestStop ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = ( 1 << StatusLed ) | ( 1 << RS232RxLed ) | ( 1 << RS232TxLed ) | ( 1 << RS485RxLed ) | ( 1 << RS485TxLed ) ;
  LEDPort |= Bits ;
} /* LEDTestStop */


int main ( void )
{
  /*----------------------------------------------------------------*/
  HardwareInit ( ) ;
  EnvirInit ( ) ;
  SoftwareInit ( ) ;
  LEDTestStart ( ) ;
  RS485UARTHardwareInit ( RS485Speed38400bps ) ;
  TL16C450ResetPort |= ( 1 << TL16C450ResetPin ) ;
  RS485ReceiveMode ( ) ;
  OpenDelay ( ) ;
  TL16C450ResetPort &= ~ ( 1 << TL16C450ResetPin ) ;
  OpenDelay ( ) ;
  RS232UARTHardwareInit ( RS232Serial9600bps ) ;
  LEDTestStop ( ) ;
  wdt_enable ( WDTO_1S ) ;
  sei ( ) ;
  RS232SendSerialFlashTxt ( ( uint16_t ) DeviceName ) ;
  RS232SendSerial ( 0x0D ) ;
  RS232SendSerial ( 0x0A ) ;
  for ( ; ; )
  {
    wdt_reset ( ) ;
    if ( TimeEvent )
    {
      TimeEvent = false ;
      LEDPort ^= ( 1 << StatusLed ) ;
    } /* if */ ;
    if ( RS232SerialDataPresent ( ) )
      RS232ProcessData ( RS232GetSerial ( ) ) ;
    if ( RS485SerialDataPresent ( ) )
      RS232SendSerial ( RS485GetSerial ( ) ) ;
  } /* for */ ;
  return ( 0 ) ;
} /* main */


